// JW

#include <cassert>
#include <stdio.h>

#include <deque>
#include <vector>
#include <algorithm>

#include <windows.h>

#include "time.h"
#include "misc.h"

#include "idx.h"


// abort reads ASAP if set
bool read_cancelled;

long cur_fn;
const char* cur_path;
unsigned cur_block;

float read_throughput;				// [byte/s]
long num_files;
u64 total_bytes, total_bytes_read;


struct File
{
	const char* name;
	const char* dir;
	u32 size;
	u32 ofs;
	FileType type;			// needed to call process_buf when getting surrounding text

	File(const char* _name, const char* _dir, u32 _size, u32 _ofs) :
	  type(FT_UNKNOWN), name(_name), dir(_dir), size(_size), ofs(_ofs) {}
};

std::vector<File> files;
static char file_list_path[MAX_PATH+1];



static bool cd;			// CD special case mode


// RAGE! MS Win32 async I/O is completely brain-dead:
// - callback order is not specified. FIFO order would help, and would
//   no trouble to implement (or am I missing something here?). fine.
// - reads do not complete in order, even if issued sequentially as given by
//   FindNextFile! (neither files, nor blocks within a file)
// - closing a file with reads in progress => undefined results. lovin' it.
// - callbacks are not only received in an alertable wait state
//   (as defined by TFM). seems they are also called while inside win calls.
//   this is fun.
//
// WHAT WERE THEY THINKING?
//
// we work around these by queuing all callbacks received after returning
// from a wait, and sorting; reads are limited until the next wait, so
// that we don't accumulate many instantly completed reads.
//
// incidentally, the easy way out - one deep reads - is a good bit slower.


// data source - either CD, or a file
struct Src
{
	int refs;

	HANDLE hFile;

	// HD only: lockstep cur_read and cur_scan counters
	// would get out of sync if opening a file fails.
	// not used in CD mode (we don't know it at read-time)
	u32 file_num;

	u32 ofs;
	u32 start, end;

	Src() {}
	Src(HANDLE hFile, u32 file_num, u32 ofs, u32 end)
	{
		this->hFile = hFile;
		this->file_num = file_num;
		this->ofs = this->start = ofs;
		this->end = end;
	}
};

struct Read
{
	OVERLAPPED ovl;	// hEvent -> this struct
	void* buf;

	u64 seq;

	Src* src;	// reading from this source; == 0 <==> this buf unused

	Read()
	{
		buf = VirtualAlloc(0, BLOCK_SIZE, MEM_COMMIT, PAGE_READWRITE);
		ovl.hEvent = this;
	}
	~Read()
	{ VirtualFree(buf, 0, MEM_DECOMMIT); }
};



//////////////////////////////////////////////////////////////////////////////
//
// Special handling for CDs
//
// We expect many small files => lots of file open and seeks.
// These are especially slow with CD drives, so we read the CD raw
// and interpret the file system ourselves.
//
// Supports ISO-9660 and Joliet CDs.
// Multisession isn't worth the trouble - we'd have to send SCSI commands
//   to the drive. Only the first session will be read.
//
//////////////////////////////////////////////////////////////////////////////

// ISO 9660 file system structures

#pragma pack(push, 1)

struct CD_DIR
{
	u8 size;
	u8 xar_sectors;
	u32 file_loc[2];		// sector #
	u32 file_size[2];
	u8 time[7];
	u8 flags;				// bit 1: dir; bit 2: associated file
	u8 _1[6];				// HACK: used as a pointer to the full path name of this entry
	u8 name_len;
	char name[1];
};

struct CD_VOL_DESC
{
	u8 type;				// PVD: 1; Joliet SVD: 2; terminator: 255
	char id[6];				// "CD001", 1
	u8 _1[33];
	char vol_id[32];
	u8 _2[8];
	u32 total_sectors[2];
	u8 _3[40];
	u16 sector_size[2];
	u32 path_tbl_size[2];
	u32 path_tbl_loc[4];	// sector #
	struct CD_DIR root_dir;
};

struct CD_PATH_TBL
{
	u8 name_len;
	u8 xar_sectors;
	u32 dir_loc;			// sector #
	u16 parent;				// index into path tbl
	char name[1];
};

#pragma pack(pop)


static Src cd_src;


// Joliet SVD present => more complete filenames available than
// Level1 interchange, but they'll need to be decoded
static bool joliet;


// we'd like to read the entire FS at once, but don't know its size
// compromise between fast list generation, and covering CDs with lots of dirs
// note: if we attempt to read FS structures not in this buffer,
//       cd_gen_file_list fails, and we revert back to standard file method
static const int FS_BUF_SIZE = 4*MB;
static void* fs_buf;
static u32 fs_bytes_read;

// return address of a sector in the fs buffer (bounds checked)
static void* sector_addr(int sector)
{
	// we haven't read this much => error
	u32 ofs = sector * 2048;
	if(ofs > fs_bytes_read)
		return 0;

	return (void*)((int)fs_buf + ofs);
}


// sort file list by sector
// (necessary for sequential file read -
//  don't want to check which file is at sector x)
static bool File_less(const File& f1, const File& f2)
{
	return f1.ofs < f2.ofs;
}


// store information about each file in path in files[]
// fails if the CD contains too many files (due to limited FS buffer)
int cd_gen_file_list(const char* path)
{
	int error = -1;		// return value - assume failure

	// open a handle to the drive
	char drive = path[0];
	static char drive_str[] = "\\\\.\\?:";
	drive_str[4] = drive;
	HANDLE hFile = CreateFile(drive_str, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_FLAG_OVERLAPPED|FILE_FLAG_NO_BUFFERING|FILE_FLAG_SEQUENTIAL_SCAN, 0);
	if(hFile == INVALID_HANDLE_VALUE)
		return -1;

	// read in first part of CD that hopefully contains the FS data
	static OVERLAPPED ovl;
	fs_buf = VirtualAlloc(0, FS_BUF_SIZE, MEM_COMMIT, PAGE_READWRITE);
	ReadFile(hFile, fs_buf, FS_BUF_SIZE, 0, &ovl);
	WaitForSingleObject(hFile, INFINITE);	// wait until read complete
	fs_bytes_read = ovl.InternalHigh;
		// note: don't wait via GetOverlappedResult for Win9x compatibility. grr.

	std::deque<CD_DIR*> dirs;
	CD_DIR* dir;

	// min/max extent of data we have to read (sector aligned)
	u32 start_ofs = ~0, end_ofs = 0;

	// get volume descriptors
	CD_VOL_DESC* pvd = (CD_VOL_DESC*)sector_addr(16);
	CD_VOL_DESC* svd = (CD_VOL_DESC*)sector_addr(17);

	// check if we read up to the PVD
	// note: 2nd volume descriptor is required (at least a terminator)
	// don't need explicit check - if we didn't read enough of it,
	// we'll fail below when accessing FS entries
	if(!pvd)
		goto fail;

	// assume ISO-9660
	joliet = false;
	dir = &pvd->root_dir;

	// it's Joliet
	if(svd && svd->type == 2)
	{
		joliet = true;
		dir = &svd->root_dir;
	}

	// this code assumes sector size = 2048, which is quite safe,
	// but not guaranteed by ISO-9660.
	if(pvd->sector_size[0] != 2048)
		return -1;

	// start dir's path with drive letter
	{
	static char root_str[] = "?:";
	root_str[0] = drive;
	*(char**)dir->_1 = root_str;
	}

	// HACK: have path test read over root dir
	path++; dir->name[0] = ':';

	dirs.push_back(dir);

	// for each dir:
	while(!dirs.empty())
	{
		dir = dirs.front(); dirs.pop_front();

		const char* dir_name = *(char**)&dir->_1;
		const int dir_name_len = strlen(dir_name);

		// skip past last path component we matched
		// if we did this below, we'd have to skip all dir entries
		// after each dir that matches (messy).
		if(*path)
			path += dir->name_len+1;

		int bytes_left = dir->file_size[0];	// of data in this dir entry

		CD_DIR* entry = (CD_DIR*)sector_addr(dir->file_loc[0]);
		if(!entry)			// FS_BUF_SIZE too small. tough.
		{
			assert(!"cd fs size > FS_BUF_SIZE");
			goto fail;
		}

		// for each dir entry:
		do
		{
			// . or ..
			if(entry->name_len == 1 && entry->name[0] <= 1)
				goto next_dir_entry;

			// convert Joliet big endian UTF-16 to ASCII (in place)
			// note: 0-terminating here doesn't help, because there might
			//       not be room in plain ISO-9660 dirents.
			if(joliet)
			{
				entry->name_len /= 2;
				for(int i = 0; i < entry->name_len; i++)
					entry->name[i] = entry->name[i*2+1];
			}

			// dir
			if(entry->flags & 2)
			{
				// make sure this dir is in the search path
				if(!*path || !strnicmp(path, entry->name, entry->name_len))
				{
					char* new_dir = (char*)mem_alloc(dir_name_len+1+entry->name_len+1, false);
					sprintf(new_dir, "%s\\%.*s", dir_name, entry->name_len, entry->name);
					*(char**)&entry->_1 = new_dir;
					dirs.push_back(entry);
				}
			}
			// file, and below search path
			// note: associated files seem to be garbage, so we skip them.
			else if(!(entry->flags & 4) && !*path)
			{
				// get rid of version, if present
				char* p = (char*)memchr(entry->name, ';', entry->name_len);
				if(p)
					entry->name_len = (p - entry->name);

				u32 ofs = entry->file_loc[0] * 2048;
				u32 size = entry->file_size[0];
				start_ofs = min(start_ofs, ofs);
				end_ofs = max(end_ofs, (u32)round_up(ofs+size, 2048));

				char* name = (char*)mem_alloc(entry->name_len+1, false);
				sprintf(name, "%.*s", entry->name_len, entry->name);
				files.push_back(File(name, dir_name, size, ofs));
			}

next_dir_entry:
			bytes_left -= entry->size;
			(unsigned&)entry += entry->size;

			// no more entries in this sector
			if(!entry->size)
			{
				u32 left_in_sector = bytes_left & 0x7ff;
				bytes_left -= left_in_sector;
				(unsigned&)entry += left_in_sector;
			}
		}
		while(bytes_left > 0);
	}

	std::sort(files.begin(), files.end(), File_less);

	num_files = files.size();
	if(num_files > 0)
	{
		total_bytes = end_ofs - start_ofs;

		cd_src.hFile = hFile;
		cd_src.ofs = cd_src.start = start_ofs;
		cd_src.end = end_ofs;

		error = 0;		// success
	}
	// no files found
	else
		error = 1;

fail:
	VirtualFree(fs_buf, 0, MEM_RELEASE);
	fs_buf = 0;

	return error;
}


static Src* cd_prepare_src()
{
	if(cd_src.ofs >= cd_src.end)
		return 0;

	return &cd_src;
}


static void cd_read_complete(const Read* read)
{
	const u8* buf = (u8*)read->buf;
	const u32 bytes_read = read->ovl.InternalHigh;

	total_bytes_read = read->ovl.Offset+bytes_read - cd_src.start;

	u32 pos = 0;
	for(;;)
	{
		File& file = files[cur_fn];
		u32 size = file.size;

		cur_block = 0;
		cur_path = file.dir;

		// where does file start, WRT current position in block?
		long delta = file.ofs - (read->ovl.Offset+pos);
		// .. in previous block
		if(delta < 0)
		{
			size += delta;
			cur_block = -delta / BLOCK_SIZE;
		}
		else
		{
			pos += delta;

			// special case: it starts far enough away that it's faster to seek there
			long skip = file.ofs - cd_src.ofs;
			if(skip >= 1048576)
				read->src->ofs = file.ofs & ~(BLOCK_SIZE-1);		// note: files are already sector aligned
		}

		long bytes_left = bytes_read - pos;
		if(bytes_left <= 0)
			break;

		bool done_with_file = false;

		if(size > (u32)bytes_left)
			size = bytes_left;
		else
			done_with_file = true;

		const u8* p = buf+pos;

		if(file.type == FT_UNKNOWN)
			file.type = detect_file_type(p, size);

		// possibly text
		if(file.type > 0)
			process_buf(p, size, file.type);
		else
			done_with_file = true;

		if(done_with_file)
			cur_fn++;

		pos += size;
	}
}


//////////////////////////////////////////////////////////////////////////////
//
// plain file version
//
//////////////////////////////////////////////////////////////////////////////

static long hd_cur_read_file_num;

// store information about each file in path in files[]
static int hd_gen_file_list(const char* path)
{
	hd_cur_read_file_num = -1;


	std::deque<const char*> dirs;		// FIFO (slightly faster)
	dirs.push_back(path);

	// for each dir:
	while(!dirs.empty())
	{
		const char* dir = dirs.front(); dirs.pop_front();
                const int dir_len = strlen(dir);

		char search_path[MAX_PATH];
		sprintf(search_path, "%s\\*", dir);

		WIN32_FIND_DATA fd;
		HANDLE hFind = FindFirstFile(search_path, &fd);
		if(hFind == INVALID_HANDLE_VALUE)
			continue;

		// for each dir entry:
		do
		{
			int name_len = strlen(fd.cFileName);

			// dir
			u32 attr = fd.dwFileAttributes, hs = FILE_ATTRIBUTE_SYSTEM|FILE_ATTRIBUTE_HIDDEN;
			if(attr & FILE_ATTRIBUTE_DIRECTORY)
			{
				// add to list if not . or .. (avoid cycle)
				// also skip hidden system dirs (recycler)
				if((fd.cFileName[0] != '.' || (fd.cFileName[1] != 0 && (fd.cFileName[1] != '.' || fd.cFileName[2] != 0))) &&
				   (attr & hs) != hs)
				{
					const char* new_dir = (char*)mem_alloc(dir_len+1+name_len+2, false);
					sprintf((char*)new_dir, "%s\\%s", dir, fd.cFileName);
					dirs.push_back(new_dir);
				}
			}
			// file
			else
			{
				const char* name = (const char*)mem_alloc(name_len+2, false);
				strcpy((char*)name, fd.cFileName);
				files.push_back(File(name, dir, fd.nFileSizeLow, 0));

				total_bytes += fd.nFileSizeLow;
			}
		}
		while(FindNextFile(hFind, &fd));

		FindClose(hFind);
	}

	num_files = files.size();
	return num_files? 0 : 1;	// 1 <==> no files found
}


static Src* hd_prepare_src()
{
	static Src* src;

	// keep reading
	if(src && src->ofs < src->end)
		return src;

	// otherwise, need new file

next_file:

	// no more files left
	if(hd_cur_read_file_num == num_files-1)
		return 0;

	File* file = &files[++hd_cur_read_file_num];

	if(cur_path != file->dir)
	{
		cur_path = file->dir;
		SetCurrentDirectory(cur_path);
	}

//printf("open %d\n", hd_cur_read_file_num);

	if(!file->size)				// don't bother opening empty files
		goto next_file;

	HANDLE hFile = CreateFile(file->name, GENERIC_READ, 0, 0, OPEN_EXISTING, FILE_FLAG_OVERLAPPED|FILE_FLAG_NO_BUFFERING|FILE_FLAG_SEQUENTIAL_SCAN, 0);
	if(hFile == INVALID_HANDLE_VALUE)
		goto next_file;

	return src = new Src(hFile, hd_cur_read_file_num, 0, round_up(file->size, 512));
}



static void hd_read_complete(const Read* read)
{
	u8* buf = (u8*)read->buf;
	u32 bytes_read = read->ovl.InternalHigh;

	HANDLE& hFile = read->src->hFile;
	File& file = files[read->src->file_num];

	cur_block = read->ovl.Offset / BLOCK_SIZE;
	cur_fn = read->src->file_num;

read->src->refs--;

	// reading was cancelled previously, but this call was already queued
	if(file.type <= 0)	// not text, or punted to 2nd pass
		return;

	if(file.type == FT_UNKNOWN)
		file.type = detect_file_type(buf, bytes_read);

	total_bytes_read += bytes_read;	

	long bytes_left = files[cur_fn].size - (read->ovl.Offset + bytes_read);
	bool done_with_file = bytes_left <= 0;

	// possibly text
	if(file.type > 0)
		process_buf(buf, bytes_read, file.type);
	// not text, or handled in second pass
	else
	{
//printf("cancel fn %d\n",read->src->file_num);
		done_with_file = true;

		// other read in flight, or pending

		if(bytes_left > 0)
		{
			CancelIo(hFile);

			total_bytes_read += bytes_left;
		}
	}



	if(done_with_file)
	{
//printf("close fn %d\n", read->src->file_num);
		CloseHandle(hFile);
		read->src->ofs = ~0;	// force new source
	}
}


//////////////////////////////////////////////////////////////////////////////
//
// shared read code
//
//////////////////////////////////////////////////////////////////////////////

static void CALLBACK read_complete_cb(u32 error, u32 bytes_read, OVERLAPPED* ovl);

static int active_reads;

static u64 issue_seq, complete_seq;


static int start_reads()
{
	// until no buffers available:
	for(;;)
	{
		Src* src = (cd)? cd_prepare_src() : hd_prepare_src();
		if(!src)
			return -1;

		// find a free read slot
		static Read reads[MAX_ACTIVE_READS*2];
		int i;
		for(i = 0; i < MAX_ACTIVE_READS*2; i++)
			if(reads[i].src == 0)
				break;
		if(i == MAX_ACTIVE_READS*2)	// none found - try again after a read completes
			return 1;

		Read* read = &reads[i];
		read->ovl.Internal = read->ovl.InternalHigh = 0;	// just in case
		read->ovl.Offset = src->ofs;
		read->src = src;
		read->seq = issue_seq++;
		src->refs++;
		u32 read_size = min(src->end - src->ofs, BLOCK_SIZE);
		src->ofs += read_size;
//printf("  read ofs %d  of %d\n", read->ovl.Offset, src->end);

// apparently ReadFileEx doesn't always set error. sigh.
#ifndef NDEBUG
		SetLastError(0);
#endif

		if(ReadFileEx(src->hFile, read->buf, read_size, &read->ovl, read_complete_cb))
		{
			active_reads++;

			// call will succeed, but check for warnings
			assert(GetLastError() == 0 && "ReadFileEx warning");
		}
		else
		{
                	assert(0 && "ReadFileEx error");
//printf("read error!\n");
			read->src = 0;		// buf available
			src->ofs = ~0;		// force new source
		}
	}

	return 0;
}


// callback order unspecified - need to sort in order of issue.
std::vector<Read*> finished_reads;


static void CALLBACK read_complete_cb(u32 error, u32 bytes_read, OVERLAPPED* ovl)
// note: error => bytes_read = 0
{
	active_reads--;

	finished_reads.push_back((Read*)ovl->hEvent);
}


// call <= 1x, after a gen_file_list()
int read_files_in_list()
{
	int error = -1;

// move to HD read code
char cur_dir[MAX_PATH];
GetCurrentDirectory(MAX_PATH, cur_dir);

	read_cancelled = false;
	total_bytes_read = 0;

	cur_fn = 0;
	cur_block = 0;


char buf2[100];
sprintf(buf2, "num_files: %d", num_files);
MessageBox(0, buf2, "yay", 0);

bool done = false;
	for(;;)
	{
		u32 bytes_read = 0;

		if(start_reads() < 0)
            	done = true;

		std::vector<Read*>::iterator it = finished_reads.begin();
		while(finished_reads.size() > 0 && it != finished_reads.end())
		{
			if((*it)->seq == complete_seq)
			{
				complete_seq++;
				cd? cd_read_complete((*it)) : hd_read_complete((*it));
				if(start_reads() < 0)
                                	done = true;

				(*it)->src = 0;	// free read slot
				bytes_read += (*it)->ovl.InternalHigh;

				finished_reads.erase(it);
				it = finished_reads.begin();
			}
			else
				++it;
		}


		calc_fps();
		read_throughput = (float)fps * bytes_read;

		// note: total_bytes_read updated in CD/HD completion function
int percent_read = (int)(total_bytes_read * 100.f / total_bytes);
fprintf(stderr, "%d %.2f\n", percent_read, read_throughput*1e-3f);

		if(active_reads == 0)
			break;
		if(read_cancelled)
			goto fail;
		SleepEx(INFINITE, 1);
	}

	error = 0;

assert(finished_reads.size() == 0 && "reorder buffer not empty");
assert(done && "not done");
assert(total_bytes_read==total_bytes && "not all data read");

fail:

SetCurrentDirectory(cur_dir);



	return error;
}


int gen_file_list(const char* path)
{
	// already got the list
//	if(!strncmp(path, file_list_path, MAX_PATH))
//		return 0;
	strncpy(file_list_path, path, MAX_PATH);

	total_bytes = 0;
	files.clear();
	num_files = 0;

	static char root_str[] = "?:\\";
	root_str[0] = path[0];

	cd = true;
	if(GetDriveType(root_str) != DRIVE_CDROM || cd_gen_file_list(path) != 0)
		// note: cd_gen_file_list fails if cd contains too many files/dirs
	{
		cd = false;
		return hd_gen_file_list(path);
	}

	return 0;
}


const char* get_filename(u32 file_num)
{
	assert(file_num < num_files);

	static char buf[MAX_PATH];
	sprintf(buf, "%s\\%s", files[file_num].dir, files[file_num].name);

	return buf;
}


long read_block(u32 file_num, u32 block, const u8*& _buf, long& size)
{
	assert(file_num < num_files);

	// only open files once
	// => bit faster, but leaks one handle at exit
/*	static HANDLE hFile;
	static u32 _cur_fn = ~0;
	if(_cur_fn != file_num)
	{
		_cur_fn = file_num;
		CloseHandle(hFile);
	}
*/
char path[MAX_PATH];
File& file = files[file_num];
sprintf(path, "%s\\%s", file.dir, file.name);

HANDLE hFile = CreateFile(path, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS, 0);

	SetFilePointer(hFile, block*BLOCK_SIZE, 0, FILE_BEGIN);

	static u8* buf;
	if(!buf)
		buf = (u8*)VirtualAlloc(0, BLOCK_SIZE, MEM_COMMIT, PAGE_READWRITE);
memset(buf,0,BLOCK_SIZE);
	u32 bytes_read;
	int success = ReadFile(hFile, buf, BLOCK_SIZE, &bytes_read, 0);
CloseHandle(hFile);

if(!success || !bytes_read)
	return -1;

_buf = buf;
size = bytes_read;

		// TODO: scan rest of file if o.block == MAX_BLOCK

	return process_buf(buf, bytes_read, files[file_num].type);
}