#include <time.h>
#include <math.h>

#include <windows.h>


#include "time.h"
#include "types.h"
#include "misc.h"

/*
 * high resolution (> 1 s) timestamp [s]
 * starts at or near 0 s.
 */
double get_time()
{
	static double to_s;
	double t;

#ifdef _WIN32

	static LARGE_INTEGER start;
	LARGE_INTEGER i;

	/* init */
	if(!to_s)
	{
		QueryPerformanceFrequency(&i);
		to_s = 1.0 / i.QuadPart;

		QueryPerformanceCounter(&start);
	}

	/* convert to seconds; asm to avoid ftol */
	QueryPerformanceCounter(&i);
	__asm
	{
		fild	qword ptr [i]
		fild	qword ptr [start]
		fsubp	st(1), st(0)		/* 64 bit, can't sub directly */
		fmul	[to_s]
		fstp	[t]
	}

#else

	static struct timeval start;
	struct timeval tv;

	/* init */
	if(!to_s)
		gettimeofday(&start, 0);

	gettimeofday(&tv, 0);
	t = (tv.tv_sec - start.tv_sec) + tv.tv_usec*1e-6;

#endif

	return t;
}


/*
 * calculate fps; fairly complicated smooth algorithm:
 * - throw out single changes of > 5%
 * - average via small history buffer
 * - update final value iff the difference (% or absolute) is too great,
 *   or if the change is consistent with the trend over the last few frames.
 *
 * => little fluctuation, but quick reaction and update time.
 */

int fps = 0;

void calc_fps()
{
	/* fps history buffer - smooth out slight variations */
#define H 10					/* # buffer entries */
	static float fps_sum = 0;	/* sum of last H frames' cur_fps */
	static float fps_hist[H];	/* last H frames' cur_fps
								   => don't need to re-average every time */
	static uint head = 0;		/* oldest entry in fps_hist
								   must be unsigned, b/c we do (head-1)%H */


	/*
	 * get elapsed time [s] since last frame;
	 * use it to approximate current fps
	 */
	static double last_t = -1.0;
	double t = get_time();

	float cur_fps = 30.0f;				/* reasonable value for first time
										   => history converges faster */
	if(last_t > 0)
		cur_fps = 1.0f / (float)(t-last_t);	/* = 1 / elapsed time */

	last_t = t;


	/*
	 * calculate fps activity over 3 frames (used below to prevent fluctuation)
	 * -1: decreasing, +1: increasing, 0: neither or fluctuating
	 */
	float h1 = fps_hist[(head-1)%H];	/* last frame's cur_fps */
	float h2 = fps_hist[(head-2)%H];	/* 2nd most recent frame's cur_fps */

	int trend = 0;
	if(h2 > h1 && h1 > cur_fps)			/* \ */
		trend = -1;
	else if(cur_fps < h1 && h1 < h2)	/* / */
		trend = 1;


	/*
	 * ignore onetime skips in fps (probably page faults or similar)
	 */
	static int bad = 0;	/* bad > 0 <==> last value was skipped */
	if(fabs(h1-cur_fps) > .05f*h1)
	{
		/*
		 * if this is the first time a bad value is encountered
		 * (bad == 0), don't update fps_hist and fps; otherwise set bad > 0
		 */
		if(!bad++)
			return;
	}
	else
		bad = 0;


	/*
	 * remove oldest cur_fps value in fps_hist from the sum
	 * and add cur_fps; also insert cur_fps in fps_hist
	 */
	fps_sum -=  fps_hist[head];
	fps_sum += (fps_hist[head] = cur_fps);
	head = (head+1)%H;


	/*
	 * update fps counter if update threshold is exceeded
	 */
	float avg_fps = fps_sum / H;

	if((trend > 0 && (avg_fps > fps || avg_fps-fps < -4.f)) ||	/* going up, or large drop */
	   (trend < 0 && (avg_fps < fps || avg_fps-fps > 4.f)) ||	/* going down, or large raise */
	   (fabs(fps-avg_fps) > min(5.f, 0.05f*fps)))				/* significant difference */
		fps = (int)avg_fps;
}